from rhetoric.view import ViewCallback

class UrlResolverMiddleware(object):
    def process_view(self, request, callback, callback_args, callback_kwargs):
        if not isinstance(callback, ViewCallback):
            return

        callback.resolve_actual_view_callable(request, callback_args, callback_kwargs)



from django.http import HttpResponse
from django.core.serializers.json import DjangoJSONEncoder


json_encode = DjangoJSONEncoder().encode


class JsonRendererFactory(object):
    def __init__(self, name):
        """ Constructor: info will be an object having the the
        following attributes: name (the renderer name), package
        (the package that was 'current' at the time the
        renderer was registered), type (the renderer type
        name), registry (the current application registry) and
        settings (the deployment settings dictionary).  """
        self.name = name

    def __call__(self, value, system=None):
        """ Call a the renderer implementation with the value
        and the system value passed in as arguments and return
        the result (a string or unicode object).  The value is
        the return value of a view.  The system value is a
        dictionary containing available system values
        (e.g. view, context, and request). """
        return HttpResponse(json_encode(value), content_type='application/json; charset=utf-8')


DEFAULT_RENDERERS = {
    'json': JsonRendererFactory,
    'string': lambda data: HttpResponse(data,
                                        content_type='text/plain; charset=utf-8'),
}


class RenderingConfiguratorMixin(object):

    def add_renderer(self, name, factory):
        self.renderers[name] = factory

    def get_renderer(self, name):
        if '.' in name:
            # template-based renderer
            _unused_part_, renderer_name = name.rsplit('.', 1)
        else:
            renderer_name = name

        try:
            return self.renderers[renderer_name](name)
        except KeyError:
            raise ValueError('No such renderer factory {}'.format(renderer_name))




class RoutesConfiguratorMixin(object):
    def add_route(self, name, pattern, rules=None, extra_kwargs=None):
        self.routes[name] = {
            'name': name,
            'pattern': pattern,
            'rules': rules,
            'extra_kwargs': extra_kwargs,
            'viewlist': [],
        }

    def get_route(self, name):
        return self.routes[name]



from django.core.urlresolvers import ResolverMatch, LocaleRegexProvider
from django.http import HttpResponse, Http404

import venusian


class view_config(object):
    venusian = venusian

    def __init__(self, **settings):
        self.__dict__.update(settings)

    def __call__(self, wrapped):
        settings = self.__dict__.copy()
        depth = settings.pop('_depth', 0)

        def callback(scanner, name, ob):
            wrapped.csrf_exempt = settings.pop('csrf_exempt', True)
            scanner.config.add_view(view=wrapped, **settings)

        self.venusian.attach(wrapped, callback, category='rhetoric', depth=depth + 1)
        return wrapped


class ViewCallback(object):
    def __init__(self, viewlist):
        self.viewlist = viewlist
        self.callback = None
        self.view_settings = None
        self.csrf_exempt = False

    def resolve_actual_view_callable(self, request, args, kwargs):
        for view_settings in self.viewlist:
            if self.check_predicates(view_settings['predicates'], request, args, kwargs):
                self.callback = view_settings['view']
                self.view_settings = view_settings
                self.csrf_exempt = self.callback.csrf_exempt
                return
        raise Http404

    def __call__(self, request, *args, **kwargs):
        response = self.callback(request, *args, **kwargs)
        return self.process_callback_response(response, self.view_settings)

    def check_predicates(self, predicates, request, req_args, req_kw):
        if request.method not in predicates['request_method']:
            return False
        return True

    def process_callback_response(self, response, view_settings):
        if isinstance(response, HttpResponse):
            # Do not process standard django responses
            return response
        renderer = view_settings['renderer']
        return renderer(response)


class RegexURLPattern(LocaleRegexProvider):
    def __init__(self, regex, default_args=None, name=None, viewlist=None):
        LocaleRegexProvider.__init__(self, regex)
        self.default_args = default_args or {}
        self.name = name
        self.viewlist = viewlist


    def resolve(self, path):
        match = self.regex.search(path)
        if match:
            # If there are any named groups, use those as kwargs, ignoring
            # non-named groups. Otherwise, pass all non-named arguments as
            # positional arguments.
            kwargs = match.groupdict()
            if kwargs:
                args = ()
            else:
                args = match.groups()
            # In both cases, pass any extra_kwargs as **kwargs.
            kwargs.update(self.default_args)

            callback = ViewCallback(self.viewlist)
            return ResolverMatch(callback, args, kwargs, self.name)



import re

from rhetoric.view import RegexURLPattern
from django.core.urlresolvers import reverse


#A replacement marker in a pattern must begin with an uppercase or
# lowercase ASCII letter or an underscore, and can be composed only
# of uppercase or lowercase ASCII letters, underscores, and numbers.
# For example: a, a_b, _b, and b9 are all valid replacement marker names, but 0a is not.
ROUTE_PATTERN_OPEN_BRACES_RE = re.compile('(?P<start_brace>\{).*')
ROUTE_PATTERN_CLOSING_BRACES_RE = re.compile('\}.*')
QUOTES_RE = re.compile('(?P<quote_type>\'\'\'|"""|\'|").*') # order matters!


def search_quotes(line, escape_char='\\', quotes_re=QUOTES_RE):
    """
    This function is taken from Plim package: https://pypi.python.org/pypi/Plim/

    ``line`` may be empty

    :param line:
    :param escape_char:
    """
    match = quotes_re.match(line)
    if not match: return None

    find_seq = match.group('quote_type')
    find_seq_len = len(find_seq)
    pos = find_seq_len
    line_len = len(line)

    while pos < line_len:
        if line[pos] == escape_char:
            pos += 2
            continue
        if line[pos:].startswith(find_seq):
            return pos + find_seq_len
        pos += 1
    return None


def _extract_braces_expression(line, starting_braces_re, open_braces_re, closing_braces_re):
    """
    This function is taken from Plim package: https://pypi.python.org/pypi/Plim/

    :param line: may be empty
    :type line: str
    :param starting_braces_re:
    :param open_braces_re:
    :param closing_braces_re:
    """
    match = starting_braces_re.match(line)
    if not match:
        return None

    open_brace = match.group('start_brace')
    buf = [open_brace]
    tail = line[len(open_brace):]
    braces_counter = 1

    while tail:
        current_char = tail[0]
        if closing_braces_re.match(current_char):
            braces_counter -= 1
            buf.append(current_char)
            if braces_counter:
                tail = tail[1:]
                continue
            return u''.join(buf), tail[1:]

        if open_braces_re.match(current_char):
            braces_counter += 1
            buf.append(current_char)
            tail = tail[1:]
            continue

        result = search_quotes(tail)
        if result is not None:
            buf.append(tail[:result])
            tail = tail[result:]
            continue

        buf.append(current_char)
        tail = tail[1:]
    raise Exception("Unexpected end of a route definition: {}".format(line))


extract_pattern = lambda line: _extract_braces_expression(
    line,
    ROUTE_PATTERN_OPEN_BRACES_RE,
    ROUTE_PATTERN_OPEN_BRACES_RE,
    ROUTE_PATTERN_CLOSING_BRACES_RE
)


def create_django_route(name, pattern, rules=None, extra_kwargs=None, viewlist=None):
    if rules is None:
        rules = {}
    buf = []
    while pattern:
        result = extract_pattern(pattern)
        if result:
            result, pattern = result
            # Remove braces from the result "{pattern[:rule]}"
            result = result[1:-1]
            if ':' in result:
                # pattern in a "pattern_name:rule" form
                match_group_name, rule = result.split(':', 1)
            else:
                # pattern in a "pattern_name" form
                match_group_name = result
                rule = rules.get(match_group_name)
                if not rule:
                    # Use a simple greedy regular expression matching everything
                    rule = '.*'

            result = u"(?P<{match_group_name}>{rule})".format(
                match_group_name=match_group_name,
                rule=rule
            )
            buf.append(result)
            continue

        buf.append(pattern[0])
        pattern = pattern[1:]

    # Parsing is done. Now join everything together
    buf = u''.join(buf)

    regex_pattern = u'^{expr}$'.format(expr=buf)
    return RegexURLPattern(regex_pattern, extra_kwargs, name, viewlist=viewlist)


def route_path(route_name, *elements, **kw):
    return reverse(route_name, args=elements, kwargs=kw)



from rhetoric.exceptions import ConfigurationError


class ViewsConfiguratorMixin(object):
    def add_view(self, view, route_name, renderer=None, request_method=None):
        try:
            route = self.routes[route_name]
        except KeyError:
            raise ConfigurationError(
                'No route named {route_name} found for view registration'.format(route_name=route_name)
            )
        route_item = {
            'view': view
        }

        if renderer is None:
            renderer = ''
        route_item['renderer'] = self.get_renderer(renderer)

        # Register predicates
        # -------------------------------------
        if request_method is None:
            request_method = {'GET'}
        elif isinstance(request_method, str):
            request_method = {request_method}
        request_method = set(request_method)

        route_item['predicates'] = {
            'request_method': request_method
        }
        route['viewlist'].append(route_item)



from collections import OrderedDict

import venusian

from rhetoric.config.rendering import RenderingConfiguratorMixin
from rhetoric.config.rendering import DEFAULT_RENDERERS
from rhetoric.config.routes import RoutesConfiguratorMixin
from rhetoric.config.views import ViewsConfiguratorMixin
from rhetoric.path import caller_package
from rhetoric.exceptions import ConfigurationError
from rhetoric.view import view_config
from rhetoric.url import route_path, create_django_route


__all__ = ['route_path', 'view_config', 'Configurator']


class Configurator(
    RoutesConfiguratorMixin,
    ViewsConfiguratorMixin,
    RenderingConfiguratorMixin):

    venusian = venusian

    def __init__(self):
        self.routes = OrderedDict()
        self.renderers = {}

        self.setup_registry()

    def django_urls(self):
        """Converts registered routes to a list of Django URLs"""
        return [
            create_django_route(
                name=route_dict['name'],
                pattern=route_dict['pattern'],
                rules=route_dict.get('rules'),
                extra_kwargs=route_dict.get('extra_kwargs'),
                viewlist=route_dict['viewlist']
            )
            for route_dict in self.routes.values()
        ]

    def scan(self, package=None, categories=None, onerror=None, ignore=None):
        if package is None:
            package = caller_package()

        scanner = self.venusian.Scanner(config=self)
        scanner.scan(package, categories=categories, onerror=onerror, ignore=ignore)
        self.check_consistency()

    def setup_registry(self):
        # Add default renderers
        # ---------------------
        for name, renderer in DEFAULT_RENDERERS.items():
            self.add_renderer(name, renderer)

    def check_consistency(self):
        for route_name, route in self.routes.items():
            viewlist = route['viewlist']
            if not viewlist:
                raise ConfigurationError(
                    'Route name "{name}" is not associated with a view callable.'.format(name=route_name)
                )
            for route_item in viewlist:
                if route_item.get('view') is None:
                    raise ConfigurationError(
                        'Route name "{name}" is not associated with a view callable.'.format(name=route_name)
                    )
